' CORRELATED MULTIVARIATE RANDOM NORMAL SCORES GENERATOR
' This program functions by first reducing a specified symmetric correlation
' matrix R into the product of a lower triangular matrix C and its
' transpose, R = C'C.  Then, a random normal vector Y ~ N(0, I) is generated
' using the Box and Muller (1958) algorithms.  Finally, the vector X is
' obtained by multiplying the generated vector Y by the lower triangular
' matrix C (i.e., X = YC ~ N(0, R)).  The computation of C for a given R is
' achieved by means of the Square Root Method (e.g., Scheuer and Stoller,
' 1962).
'
'Citation:
'  Aguinis, H. (1994).  A QuickBASIC program for generating correlated
'    multivariate random normal scores.  Educational and Psychological
'    Measurement, 54, 687-689.
'
' E-mail: haguinis@gwu.edu
' WWW: https://www.hermanaguinis.com

'************************************************************************
DECLARE SUB DisplayHeader ()
DECLARE SUB InitializeVariables ()
DECLARE SUB WriteStats (Sample!, NumberOfSamples!, N!, NumberOfCases!)
DECLARE SUB WriteCaseLowAcc (N!, CaseNumber!)
DECLARE SUB WriteCaseHighAcc (N!, CaseNumber!)
DECLARE SUB DisplayStats (Sample!, NumberOfSamples!, N!, NumberOfCases!)
DECLARE SUB ComputeStats (N!, NumberOfCases!)
DECLARE SUB GenerateXvalues (N!, NumberOfCases!)
DECLARE SUB GenerateMultivariate (N!, NumberOfCases!)
DECLARE SUB GenerateMatrixC (N!)
DECLARE SUB GetCovariances (N!)
DECLARE SUB GetUserInformation (N!, NumberOfCases!, NumberOfSamples!)
DECLARE SUB GenerateYvalues (N!)

'************************************************************************
'Array Declarations
'
'************************************************************************
Dim Shared Covar!(10, 10) 'Covariance relations matrix
Dim Shared c!(10, 10) 'The matrix C
Dim Shared X!(10, 1000) 'The X vector
Dim Shared y!(10) 'The Y vector

'Arrays needed to compute statistics for the generated data
Dim Shared SumX!(10)
Dim Shared SumXsquare!(10)
Dim Shared SumXcross!(10, 10)
Dim Shared Xbar!(10)
Dim Shared SumXdevSqr!(10)
Dim Shared StdDevOfX!(10)
Dim Shared CorrX!(10, 10)

'************************************************************************
'  Main Body of Program
'
'************************************************************************
Cls 'Clear the screen
Randomize Timer

Call DisplayHeader
Call GetUserInformation(N, NumberOfCases, NumberOfSamples)
Call GetCovariances(N)
Call GenerateMatrixC(N)
Locate 1, 60
Color 31
Print " *** Please wait *** "
Color 7

For Sample = 1 To NumberOfSamples
    Call InitializeVariables
    Call GenerateMultivariate(N, NumberOfCases)
    Call ComputeStats(N, NumberOfCases)
    Call DisplayStats(Sample, NumberOfSamples, N, NumberOfCases)
    Call WriteStats(Sample, NumberOfSamples, N, NumberOfCases)

Next Sample

Locate 1, 60: Print "                    "
End 'Program

'************************************************************************
'Compute Statistics Module
'
'    This module computes the sample-based means, standard deviations,
'    and correlations
'
'
'    Variables Used in Module
'    ------------------------
'         i, j                    Local to Module
'         N, NumberOfCases        Parameters to Module
'         CorrX ()                Global to Module
'         StdDevOfX ()            Global to Module
'         SumX ()                 Global to Module
'         SumXcross ()            Global to Module
'         SumXdevSqr ()           Global to Module
'         SumXsquare()            Global to Module
'         Xbar ()                 Global to Module
'
'    Variables Modified by this Module
'    ---------------------------------
'         CorrX ()
'         StdDevOfX ()
'         SumX ()
'         SumXcross ()
'         SumXdevSqr ()
'         SumXsquare()
'
'************************************************************************
Sub ComputeStats (N, NumberOfCases)

    For CaseNumber = 1 To NumberOfCases
        For i = 1 To N
            SumX(i) = SumX(i) + X(i, CaseNumber)
            SumXsquare(i) = SumXsquare(i) + X(i, CaseNumber) ^ 2
            For j = (i + 1) To N
                SumXcross(i, j) = SumXcross(i, j) + (X(i, CaseNumber) * X(j, CaseNumber))

            Next j
        Next i
    Next CaseNumber

    For i = 1 To N
        Xbar(i) = SumX(i) / NumberOfCases
    Next i

    'Initialize SumXdevSqr array
    For i = 1 To N: SumXdevSqr(i) = 0!: Next i

    For CaseNumber = 1 To NumberOfCases
        For i = 1 To N
            SumXdevSqr(i) = (X(i, CaseNumber) - Xbar(i)) ^ 2 + SumXdevSqr(i)
        Next i
    Next CaseNumber

    For i = 1 To N
        StdDevOfX(i) = Sqr(SumXdevSqr(i) / (NumberOfCases - 1))
    Next i

    For i = 1 To N
        For j = (i + 1) To N
            CorrX(i, j) = (NumberOfCases * SumXcross(i, j) - SumX(i) * SumX(j)) / (Sqr(NumberOfCases * SumXsquare(i) - SumX(i) ^ 2) * Sqr(NumberOfCases * SumXsquare(j) - SumX(j) ^ 2))
        Next j
    Next i



End Sub

Sub DisplayHeader
    Print: Print: Print
    Print: Print
    Print "           CORRELATED MULTIVARIATE RANDOM NORMAL SCORES GENERATOR ";
    Print
    Print "                              Herman Aguinis"
    Print "                    The George Washington University"
    Print: Print "                 https://www.hermanaguinis.com"
    Print "                 email: haguinis@gwu.edu"
    Print: Print
    Print "        Citation:"
    Print "        Aguinis, H. (1994).  A QuickBASIC program for generating correlated"
    Print "        multivariate random normal scores.  Educational and Psychological"
    Print "        Measurement, 54, 687-689."
    Print ""
    Print: Print: Print: Print: Print: Print
    Print " Press any key to continue "
    Print
    ak$ = Input$(1)
    Cls
End Sub

'************************************************************************
'Display Statistics Module
'
'    This module displays the statistics as they are being computed.
'
'    Variables Used in Module
'    ------------------------
'         i, j                     Local to Module
'         N, NumberOfCases         Parameters to Module
'         StdDevOfX ()             Global to Module
'         Xbar ()                  Global to Module
'         CorrX ()                 Global to Module
'
'    Variables Modified by this Module
'    ---------------------------------
'         none
'
'************************************************************************
Sub DisplayStats (Sample, NumberOfSamples, N, NumberOfCases)
    Cls
    Locate 1, 60
    Color 31
    Print " *** Please wait *** "
    Color 7

    Locate 2, 1
    Print "Sample"; Sample; " of "; NumberOfSamples
    Print "Sample Size = "; NumberOfCases

    For i = 1 To N
        Print "Mean X"; i; " = ";: Print Using "##.######"; Xbar(i)
    Next i

    For i = 1 To N
        Print "Standard Deviation X"; i; " = "; StdDevOfX(i)
    Next i

    For i = 1 To N
        For j = (i + 1) To N
            Print "r X"; i; "X"; j; " = ";: Print Using "##.######"; CorrX(i, j)
        Next j
    Next i

End Sub

'************************************************************************
'Generate Matrix C Module
'
'    This module creates the matrix C via the Square Root Method
'
'
'    Variables Used in Module
'    ------------------------
'         i, j, k                       Local to Module
'         Csum                          Local to Module
'         N                             Parameter to Module
'         C ()                          Global to Module
'         Covar ()                      Global to Module
'
'    Variables Modified by this Module
'    ---------------------------------
'         C ()
'
'************************************************************************
Sub GenerateMatrixC (N)
    For i = 1 To N
        For j = 1 To N

            'CASE #1
            If (j = 1) Then
                c(i, j) = Covar(i, 1) / Sqr(Covar(1, 1))

                'CASE #2
            ElseIf ((i = j) And (i <> 1)) Then
                Csum = 0

                For k = 1 To (i - 1)
                    Csum = Csum + c(i, k) ^ 2

                Next k

                c(i, j) = Sqr(Covar(i, j) - Csum)

                'CASE #3
            ElseIf ((j < i) And (i <> 1) And (j <> 1)) Then
                Csum = 0

                For k = 1 To (j - 1)
                    Csum = Csum + c(i, k) * c(j, k)

                Next k

                c(i, j) = (Covar(i, j) - Csum) / c(j, j)

                'CASE #4
            Else
                c(i, j) = 0!

            End If

        Next j

    Next i

End Sub

'************************************************************************
'Generate Multivariate Subroutine
'
'    This module controls the sequence for caculating the necessary
'    values for the matrix C, the Y vector, and the X vector.
'
'
'    Variables Used in Module
'    ------------------------
'         NumberOfCases
'
'
'    Variables Modified by the Module
'    --------------------------------
'         none
'
'************************************************************************
Sub GenerateMultivariate (N, NumberOfCases)

    For CaseNumber = 1 To NumberOfCases
        Call GenerateYvalues(N)
        Call GenerateXvalues(N, CaseNumber)

        If N <= 7 Then 'Number of variables 1-7 write higher accuracy
            Call WriteCaseHighAcc(N, CaseNumber)

        Else 'Number of variables 7-10 write lower accuracy
            Call WriteCaseLowAcc(N, CaseNumber)

        End If

    Next CaseNumber

End Sub

'************************************************************************
'Generate X Vector Module
'
'    This module generates the X vector for a given CaseNumber.
'
'    Variables Used in Module
'    ------------------------
'         i, j                          Local to Module
'         N, CaseNumber                 Parameter to Module
'         C ()                          Global to Module
'         Covar ()                      Global to Module
'         X ()                          Global to Module
'
'    Variables Modified by this Module
'    ---------------------------------
'         X ()
'
'************************************************************************
Sub GenerateXvalues (N, CaseNumber)
    For i = 1 To N
        X(i, CaseNumber) = 0!

        For j = 1 To i
            X(i, CaseNumber) = X(i, CaseNumber) + c(i, j) * y(j)

        Next j

    Next i

End Sub

'************************************************************************
'Generate Y Vector Module
'
'    This module generates the Y vector to be used in the calculation
'    of the X vector.
'
'    Variables Used in Module
'    ------------------------
'         i                             Local to Module
'         U1, U2                        Local to Module
'         pi                            Local Constant to Module
'         N                             Parameter to Module
'         Y ()                          Global to Module
'
'    Variables Modified by this Module
'    ---------------------------------
'         Y ()
'
'************************************************************************
Sub GenerateYvalues (N)
    pi = 3.1415926535898#

    i = 1
    Do While i <= N
        U1 = Rnd
        U2 = Rnd


        y(i) = Sqr(-2! * (Log(U1))) * Cos(2! * pi * U2)
        i = i + 1

        If i <= N Then
            y(i) = Sqr(-2! * (Log(U1))) * Sin(2! * pi * U2)
        Else:
            i = i + 1

        End If

    Loop
End Sub

'************************************************************************
'Get Covariances Module
'
'    This module gets the population covariances (correlations in this case)
'    from the user.
'
'    Variables Used in Module
'    ------------------------
'         i, j                          Local to Module
'         N                             Parameter to Module
'         Covar ()                      Global to Module
'
'    Variables Modified by the Module
'    --------------------------------
'         Covar ()
'
'************************************************************************
Sub GetCovariances (N)

    i = 1
    Do While i <= N
        j = i
        Covar(i, j) = 1
        j = j + 1

        Do While j <= N


            c: Print "Population Correlation between X"; i; " and  X"; j; ":  ";
            Input Covar(i, j)
            If Covar(i, j) > 1! Or Covar(i, j) < -1! Then GoTo c:
            Covar(j, i) = Covar(i, j)
            j = j + 1

        Loop 'End DO WHILE j <= N

        i = i + 1

    Loop 'End DO WHILE i <= N

End Sub

'************************************************************************
'Get User Information Module
'
'    This module gets the number of variables (N), the
'    number of cases (NumberOfCases), and the number of samples (NumberOfSamples)
'
'    Variables Used in Module
'    ------------------------
'         N, NumberOfCases              Parameters to Module
'
'    Variables Modified by the Module
'    --------------------------------
'         N, NumberOfCases
'
'************************************************************************
Sub GetUserInformation (N, NumberOfCases, NumberOfSamples)
    Locate 3, 1
    Input "Enter the number of samples  :  "; NumberOfSamples
    a: Input "Enter the number of variables:  "; N
    If N > 10 Then GoTo a:
    b: Input "Enter the number of cases:      "; NumberOfCases
    If NumberOfCases > 1000 Then GoTo b:
    Print

End Sub

Sub InitializeVariables

    For i = 1 To 10
        SumX!(i) = 0!
        SumXsquare!(i) = 0!
        Xbar!(i) = 0!
        SumXdevSqr!(i) = 0!
        StdDevOfX!(i) = 0!
        y!(i) = 0!

        For j = 1 To 10
            SumXcross!(i, j) = 0!
            CorrX!(i, j) = 0!
        Next j

    Next i

    For i = 1 To 10
        For j = 1 To 1000
            X!(i, j) = 0!
        Next j
    Next i

End Sub

'************************************************************************
' Write Case to File with High Accuracy Module
'
'    This module writes out the given CASE for a generated X vector.
'    The values written are to 7 decimal places.
'
'    Variables Used in Module
'    ------------------------
'         i                             Local to module
'         CaseNumber                    Parameter to module
'         N                             Parameter to module
'         X                             Global to module
'
'    Variables Modified by the Module
'    --------------------------------
'         none
'
'************************************************************************
Sub WriteCaseHighAcc (N, CaseNumber)
    Open "data.out" For Append As #1

    For i = 1 To N
        Print #1, Using "##.#######"; X(i, CaseNumber);
        Print #1, " ";
    Next i

    Print #1, ""

    Close #1

End Sub

'************************************************************************
' Write Case to File with Low Accuracy Module
'
'    This module writes out the given CASE for a generated X vector.
'    The values written are to 3 decimal places.
'
'    Variables Used in Module
'    ------------------------
'         i                             Local to module
'         CaseNumber                    Parameter to module
'         N                             Parameter to module
'         X                             Global to module
'
'    Variables Modified by the Module
'    --------------------------------
'         none
'
'************************************************************************
Sub WriteCaseLowAcc (N, CaseNumber)
    Open "data.out" For Append As #1

    For i = 1 To N
        Print #1, Using "##.###"; X(i, CaseNumber);
        Print #1, " ";
    Next i

    Print #1, ""

    Close #1

End Sub

'************************************************************************
' Write Statistics to File Module
'
'    This module writes the statistics out each sample
'
'    Variables Used in Module
'    ------------------------
'         i                             Local to module
'         j                             Local to module
'         N                             Parameter to module
'         NumberOfCases                 Parameter to module
'         NumberOfSamples               Parameter to module
'         Sample                        Parameter to module
'         CorrX                         Global to module
'         StdDevOfX                     Global to module
'         Xbar                          Global to module
'
'    Variables Modified by the Module
'    --------------------------------
'         none
'
'************************************************************************
Sub WriteStats (Sample, NumberOfSamples, N, NumberOfCases)
    Open "stats.out" For Append As #1
    Print #1, "Sample "; Sample; " of "; NumberOfSamples
    Print #1, "Sample Size = "; NumberOfCases
    Print #1, ""

    For i = 1 To N
        Print #1, "Mean X"; i; " = ";
        Print #1, Using "#.#######"; Xbar(i)
    Next i
    Print #1, ""

    For i = 1 To N
        Print #1, "Standard Deviation X"; i; " = "; StdDevOfX(i)
    Next i
    Print #1, ""

    For i = 1 To N
        For j = (i + 1) To N
            Print #1, "r X"; i; "X"; j; " = ";
            Print #1, Using "#.#######"; CorrX(i, j)
        Next j
    Next i

    Print #1, ""
    Print #1, "*******************************************************"
    Print #1, ""
    Close #1

End Sub


